@@ -7,7 +7,7 @@
   It assumes we are running as 32 bit int with a 64 bit long long type
   available.
 
-  Copyright (C) 2004-2005 Roger Binns <rogerb@rogerbinns.com>
+  Copyright (C) 2004-2006 Roger Binns <rogerb@rogerbinns.com>
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any
@@ -38,6 +38,12 @@
 #include <pythread.h>
 #include "structmember.h"
 
+/* Python 2.5 compatibility when size_t types become 64 bit.
+   SQLite3 is limited to 32 bit sizes even on a 64 bit machine. */
+#if PY_VERSION_HEX < 0x02050000
+typedef int Py_ssize_t;
+#endif
+
 /* SQLite 3 headers */
 #include "sqlite3.h"
 
@@ -54,7 +60,7 @@
    affects how strings are stored in the database.  We use utf8 since
    it is more space efficient, and Python can't make its mind up about
    Unicode (it uses 16 or 32 bit unichars and often likes to use Byte
-   Order Markers. */
+   Order Markers). */
 #define STRENCODING "utf_8"
 
 /* Some macros used for frequent operations */
@@ -80,6 +86,7 @@ static PyObject *ExcIncomplete;  /* didn't finish previous query */
 static PyObject *ExcBindings;  /* wrong number of bindings */
 static PyObject *ExcComplete;  /* query is finished */
 static PyObject *ExcTraceAbort; /* aborted by exectrace */
+static PyObject *ExcTooBig; /* object is too large for SQLite */
 
 static struct { int code; const char *name; PyObject *cls;}
 exc_descriptors[]=
@@ -154,6 +161,7 @@ static int init_exceptions(PyObject *m)
   EXC(ExcBindings, "BindingsError");
   EXC(ExcComplete, "ExecutionCompleteError");
   EXC(ExcTraceAbort, "ExecTraceAbort");
+  EXC(ExcTooBig, "TooBigError");
 
 #undef EXC
 
@@ -181,12 +189,12 @@ static void make_exception(int res, sqlite3 *db)
     if (exc_descriptors[i].code==res)
       {
         assert(exc_descriptors[i].cls);
-        PyErr_Format(exc_descriptors[i].cls, "%sError: %s", exc_descriptors[i].name, sqlite3_errmsg(db));
+        PyErr_Format(exc_descriptors[i].cls, "%sError: %s", exc_descriptors[i].name, db?(sqlite3_errmsg(db)):"error");
         assert(PyErr_Occurred());
         return;
       }
 
-  PyErr_Format(APSWException, "Error %d: %s", res, sqlite3_errmsg(db));  
+  PyErr_Format(APSWException, "Error %d: %s", res, db?(sqlite3_errmsg(db)):"error");  
 }
 
 /* If res indicates an SQLite error then do all the exception creation
@@ -235,7 +243,10 @@ typedef struct {
   collationcbinfo *collations;    /* linked list of registered collations */
 
   /* registered hooks/handlers (NULL or callable) */
-  PyObject *busyhandler;         
+  PyObject *busyhandler;     
+  PyObject *rollbackhook;
+  PyObject *profile;
+  PyObject *updatehook;
   PyObject *commithook;           
   PyObject *progresshandler;      
   PyObject *authorizer;
@@ -259,7 +270,7 @@ typedef struct {
 
   /* bindings for query */
   PyObject *bindings;             /* dict or sequence */
-  int bindingsoffset;             /* for sequence tracks how far along we are when dealing with multiple statements */
+  Py_ssize_t bindingsoffset;             /* for sequence tracks how far along we are when dealing with multiple statements */
 
   /* iterator for executemany */
   PyObject *emiter;
@@ -276,7 +287,7 @@ static PyTypeObject CursorType;
 /* CONVENIENCE FUNCTIONS */
 
 /* Convert a NULL terminated UTF-8 string into a Python object.  None
-   is returned is NULL is passed in. */
+   is returned if NULL is passed in. */
 static PyObject *
 convertutf8string(const char *str)
 {
@@ -298,10 +309,10 @@ convertutf8string(const char *str)
 /* Convert a pointer and size UTF-8 string into a Python object.
    Pointer must be non-NULL. */
 static PyObject *
-convertutf8stringsize(const char *str, int size)
+convertutf8stringsize(const char *str, Py_ssize_t size)
 {
   const char *chk=str;
-  int i;
+  Py_ssize_t i;
   
   assert(str);
   assert(size>=0);
@@ -337,7 +348,7 @@ convertutf8stringsize(const char *str, int size)
 #define UNIDATABEGIN(obj) \
 {                                                        \
   const int use16=0;                                     \
-  size_t strbytes=0;                                     \
+  Py_ssize_t strbytes=0;				 \
   const char *strdata=NULL;                              \
   PyObject *_utf8=NULL;                                  \
                                                          \
@@ -417,6 +428,12 @@ Connection_dealloc(Connection* self)
   Py_XDECREF(self->busyhandler);
   self->busyhandler=0;
 
+  Py_XDECREF(self->rollbackhook);
+  self->rollbackhook=0;
+
+  Py_XDECREF(self->profile);
+  self->profile=0;
+
   Py_XDECREF(self->commithook);
   self->commithook=0;
 
@@ -444,6 +461,9 @@ Connection_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
       self->functions=0;
       self->collations=0;
       self->busyhandler=0;
+      self->rollbackhook=0;
+      self->profile=0;
+      self->updatehook=0;
       self->commithook=0;
       self->progresshandler=0;
       self->authorizer=0;
@@ -548,6 +568,15 @@ Connection_totalchanges(Connection *self)
   return Py_BuildValue("i", sqlite3_total_changes(self->db));
 }
 
+static PyObject *
+Connection_getautocommit(Connection *self)
+{
+  PyObject *res;
+  CHECK_THREAD(self,NULL);
+  res=(sqlite3_get_autocommit(self->db))?(Py_True):(Py_False);
+  Py_INCREF(res);
+  return res;
+}
 
 static PyObject *
 Connection_last_insert_rowid(Connection *self)
@@ -597,6 +626,244 @@ Connection_interrupt(Connection *self)
   return Py_BuildValue("");
 }
 
+void
+updatecb(void *context, int updatetype, char const *databasename, char const *tablename, sqlite_int64 rowid)
+{
+  /* The hook returns void. That makes it impossible for us to
+     abort immediately due to an error in the callback */
+  
+  PyGILState_STATE gilstate;
+  PyObject *retval=NULL, *args=NULL;
+  Connection *self=(Connection *)context;
+  PyObject *pupdatetype=NULL, *pdatabasename=NULL, *ptablename=NULL, *prowid=NULL;
+
+  assert(self);
+  assert(self->updatehook);
+  assert(self->updatehook!=Py_None);
+
+  /* defensive coding */
+  if(!self->updatehook)
+    return;
+
+  gilstate=PyGILState_Ensure();
+
+  if(PyErr_Occurred())
+    goto finally;  /* abort hook due to outstanding exception */
+
+
+  pupdatetype=Py_BuildValue("i", updatetype);
+  pdatabasename=convertutf8string(databasename);
+  ptablename=convertutf8string(tablename);
+  prowid=PyLong_FromLongLong(rowid);
+
+  if (!pupdatetype || !pdatabasename || !ptablename || !prowid)
+    goto finally;
+
+  args=PyTuple_New(4);
+  if(!args)
+    goto finally; /* abort hook on failure to allocate args */
+
+  PyTuple_SET_ITEM(args, 0, pupdatetype);
+  PyTuple_SET_ITEM(args, 1, pdatabasename);
+  PyTuple_SET_ITEM(args, 2, ptablename);
+  PyTuple_SET_ITEM(args, 3, prowid);
+
+  pupdatetype=pdatabasename=ptablename=prowid=NULL; /* owned by args now */
+  
+  retval=PyEval_CallObject(self->updatehook, args);
+
+ finally:
+  Py_XDECREF(retval);
+  Py_XDECREF(args);
+  Py_XDECREF(pupdatetype);
+  Py_XDECREF(pdatabasename);
+  Py_XDECREF(ptablename);
+  Py_XDECREF(prowid);
+  PyGILState_Release(gilstate);
+}
+
+static PyObject *
+Connection_setupdatehook(Connection *self, PyObject *callable)
+{
+  /* sqlite3_update_hook doesn't return an error code */
+  
+  CHECK_THREAD(self,NULL);
+
+  if(callable==Py_None)
+    {
+      sqlite3_update_hook(self->db, NULL, NULL);
+      callable=NULL;
+      goto finally;
+    }
+
+  if(!PyCallable_Check(callable))
+    {
+      PyErr_Format(PyExc_TypeError, "update hook must be callable");
+      return NULL;
+    }
+
+  sqlite3_update_hook(self->db, updatecb, self);
+
+  Py_INCREF(callable);
+
+ finally:
+
+  Py_XDECREF(self->updatehook);
+  self->updatehook=callable;
+
+  return Py_BuildValue("");
+}
+
+void
+rollbackhookcb(void *context)
+{
+  /* The hook returns void. That makes it impossible for us to
+     abort immediately due to an error in the callback */
+  
+  PyGILState_STATE gilstate;
+  PyObject *retval=NULL, *args=NULL;
+  Connection *self=(Connection *)context;
+
+  assert(self);
+  assert(self->rollbackhook);
+  assert(self->rollbackhook!=Py_None);
+
+  /* defensive coding */
+  if(!self->rollbackhook)
+    return;
+
+  gilstate=PyGILState_Ensure();
+
+  if(PyErr_Occurred())
+    goto finally;  /* abort hook due to outstanding exception */
+
+  args=PyTuple_New(0);
+  if(!args)
+    goto finally; /* abort hook on failure to allocate args */
+  
+  retval=PyEval_CallObject(self->rollbackhook, args);
+
+ finally:
+  Py_XDECREF(retval);
+  Py_XDECREF(args);
+  PyGILState_Release(gilstate);
+}
+
+static PyObject *
+Connection_setrollbackhook(Connection *self, PyObject *callable)
+{
+  /* sqlite3_rollback_hook doesn't return an error code */
+  
+  CHECK_THREAD(self,NULL);
+
+  if(callable==Py_None)
+    {
+      sqlite3_rollback_hook(self->db, NULL, NULL);
+      callable=NULL;
+      goto finally;
+    }
+
+  if(!PyCallable_Check(callable))
+    {
+      PyErr_Format(PyExc_TypeError, "rollback hook must be callable");
+      return NULL;
+    }
+
+  sqlite3_rollback_hook(self->db, rollbackhookcb, self);
+
+  Py_INCREF(callable);
+
+ finally:
+
+  Py_XDECREF(self->rollbackhook);
+  self->rollbackhook=callable;
+
+  return Py_BuildValue("");
+}
+
+#ifdef EXPERIMENTAL /* sqlite3_profile */
+void
+profilecb(void *context, const char *statement, sqlite_uint64 runtime)
+{
+  /* The hook returns void. That makes it impossible for us to
+     abort immediately due to an error in the callback */
+  
+  PyGILState_STATE gilstate;
+  PyObject *retval=NULL, *args=NULL;
+  Connection *self=(Connection *)context;
+  PyObject *pstatement=NULL, *pruntime=NULL;
+
+  assert(self);
+  assert(self->profile);
+  assert(self->profile!=Py_None);
+
+  /* defensive coding */
+  if(!self->profile)
+    return;
+
+  gilstate=PyGILState_Ensure();
+
+  if(PyErr_Occurred())
+    goto finally;  /* abort hook due to outstanding exception */
+
+  pstatement=convertutf8string(statement);
+  pruntime=PyLong_FromUnsignedLongLong(runtime);
+
+  if (!pstatement || !pruntime)
+    goto finally;
+
+  args=PyTuple_New(2);
+  if(!args)
+    goto finally; /* abort hook on failure to allocate args */
+  
+  PyTuple_SET_ITEM(args, 0, pstatement);
+  PyTuple_SET_ITEM(args, 1, pruntime);
+
+  pstatement=pruntime=NULL; /* owned by args now */
+
+  retval=PyEval_CallObject(self->profile, args);
+
+ finally:
+  Py_XDECREF(retval);
+  Py_XDECREF(args);
+  Py_XDECREF(pstatement);
+  Py_XDECREF(pruntime);
+  PyGILState_Release(gilstate);
+}
+
+static PyObject *
+Connection_setprofile(Connection *self, PyObject *callable)
+{
+  /* sqlite3_profile doesn't return an error code */
+  
+  CHECK_THREAD(self,NULL);
+
+  if(callable==Py_None)
+    {
+      sqlite3_profile(self->db, NULL, NULL);
+      callable=NULL;
+      goto finally;
+    }
+
+  if(!PyCallable_Check(callable))
+    {
+      PyErr_Format(PyExc_TypeError, "profile function must be callable");
+      return NULL;
+    }
+
+  sqlite3_profile(self->db, profilecb, self);
+
+  Py_INCREF(callable);
+
+ finally:
+
+  Py_XDECREF(self->profile);
+  self->profile=callable;
+
+  return Py_BuildValue("");
+}
+#endif /* EXPERIMENTAL - sqlite3_profile */
+
 
 #ifdef EXPERIMENTAL      /* commit hook */
 int commithookcb(void *context)
@@ -1001,12 +1268,12 @@ convert_value_to_pyobject(sqlite3_value *value)
     case SQLITE_BLOB:
       {
         PyObject *item;
-        int sz=sqlite3_value_bytes(value);
+        Py_ssize_t sz=sqlite3_value_bytes(value);
         item=PyBuffer_New(sz);
         if(item)
           {
             void *buffy=0;
-            int sz2=sz;
+            Py_ssize_t sz2=sz;
             if(!PyObject_AsWriteBuffer(item, &buffy, &sz2))
               memcpy(buffy, sqlite3_value_blob(value), sz);
             else
@@ -1050,7 +1317,7 @@ set_context_result(sqlite3_context *context, PyObject *obj)
     }
   if(PyInt_Check(obj))
     {
-      sqlite3_result_int(context, PyInt_AS_LONG(obj));
+      sqlite3_result_int64(context, PyInt_AS_LONG(obj));
       return;
     }
   if (PyLong_Check(obj))
@@ -1068,10 +1335,17 @@ set_context_result(sqlite3_context *context, PyObject *obj)
       UNIDATABEGIN(obj)
         if(strdata)
           {
-            if(use16)
-              sqlite3_result_text16(context, strdata, strbytes, SQLITE_TRANSIENT);
-            else
-              sqlite3_result_text(context, strdata, strbytes, SQLITE_TRANSIENT);
+	    if(strbytes>INT32_MAX)
+	      {
+		PyErr_Format(ExcTooBig, "Unicode object is too large - SQLite only supports up to 2GB");
+	      }
+	    else
+	      {
+		if(use16)
+		  sqlite3_result_text16(context, strdata, (int)strbytes, SQLITE_TRANSIENT);
+		else
+		  sqlite3_result_text(context, strdata, (int)strbytes, SQLITE_TRANSIENT);
+	      }
           }
         else
           sqlite3_result_error(context, "Unicode conversions failed", SQLITE_ERROR);
@@ -1094,10 +1368,17 @@ set_context_result(sqlite3_context *context, PyObject *obj)
           UNIDATABEGIN(str2)
             if(strdata)
               {
-                if(use16)
-                  sqlite3_result_text16(context, strdata, strbytes, SQLITE_TRANSIENT);
-                else
-                  sqlite3_result_text(context, strdata, strbytes, SQLITE_TRANSIENT);
+		if(strbytes>INT32_MAX)
+		  {
+		    PyErr_Format(ExcTooBig, "Unicode object is too large - SQLite only supports up to 2GB");
+		  }
+		else
+		  {
+		    if(use16)
+		      sqlite3_result_text16(context, strdata, (int)strbytes, SQLITE_TRANSIENT);
+		    else
+		      sqlite3_result_text(context, strdata, (int)strbytes, SQLITE_TRANSIENT);
+		  }
               }
             else
               sqlite3_result_error(context, "Unicode conversions failed", SQLITE_ERROR);
@@ -1105,19 +1386,30 @@ set_context_result(sqlite3_context *context, PyObject *obj)
           Py_DECREF(str2);
         }
       else
-        sqlite3_result_text(context, val, PyString_GET_SIZE(obj), SQLITE_TRANSIENT);
+	{
+	  Py_ssize_t lenval=PyString_GET_SIZE(obj);
+	  if(lenval>INT32_MAX)
+	      {
+		PyErr_Format(ExcTooBig, "String object is too large - SQLite only supports up to 2GB");
+		}
+	  else
+	    sqlite3_result_text(context, val, (int)lenval, SQLITE_TRANSIENT);
+	}
       return;
     }
   if (PyBuffer_Check(obj))
     {
       const char *buffer;
-      int buflen;
+      Py_ssize_t buflen;
       if(PyObject_AsCharBuffer(obj, &buffer, &buflen))
         {
           sqlite3_result_error(context, "PyObject_AsCharBuffer failed", SQLITE_ERROR);
           return;
         }
-      sqlite3_result_blob(context, buffer, buflen, SQLITE_TRANSIENT);
+      if (buflen>INT32_MAX)
+	sqlite3_result_error(context, "Buffer object is too large for SQLite - only up to 2GB is supported", SQLITE_ERROR);
+      else
+	sqlite3_result_blob(context, buffer, (int)buflen, SQLITE_TRANSIENT);
       return;
     }
 
@@ -1138,7 +1430,7 @@ getfunctionargs(sqlite3_context *context, PyObject *firstelement, int argc, sqli
   if(firstelement)
     extra=1;
 
-  pyargs=PyTuple_New(argc+extra);
+  pyargs=PyTuple_New((long)argc+extra);
   if(!pyargs)
     {
       sqlite3_result_error(context, "PyTuple_New failed", SQLITE_ERROR);
@@ -1171,7 +1463,7 @@ getfunctionargs(sqlite3_context *context, PyObject *firstelement, int argc, sqli
 }
 
 
-
+/* dispatches scalar function */
 static void
 cbdispatch_func(sqlite3_context *context, int argc, sqlite3_value **argv)
 {
@@ -1272,7 +1564,6 @@ getaggregatefunctioncontext(sqlite3_context *context)
   assert(retval);
   Py_DECREF(retval);
   return aggfc;
-
 }
 
 
@@ -1303,12 +1594,6 @@ cbdispatch_step(sqlite3_context *context, int argc, sqlite3_value **argv)
 
   assert(aggfc);
   
-  if(aggfc->aggvalue==Py_None)
-    {
-      assert(PyErr_Occurred());
-      goto finally;
-    }
-
   pyargs=getfunctionargs(context, aggfc->aggvalue, argc, argv);
   if(!pyargs)
     goto finally;
@@ -1333,19 +1618,23 @@ static void
 cbdispatch_final(sqlite3_context *context)
 {
   PyGILState_STATE gilstate;
-  PyObject *pyargs;
-  PyObject *retval;
+  PyObject *pyargs=NULL;
+  PyObject *retval=NULL;
   aggregatefunctioncontext *aggfc=NULL;
+  PyObject *err_type=NULL, *err_value=NULL, *err_traceback=NULL;
 
   gilstate=PyGILState_Ensure();
 
+  PyErr_Fetch(&err_type, &err_value, &err_traceback);
+  PyErr_Clear();
+
   aggfc=getaggregatefunctioncontext(context);
 
   assert(aggfc);
   
-  if(PyErr_Occurred() || !aggfc->finalfunc)
+  if((err_type||err_value||err_traceback) || PyErr_Occurred() || !aggfc->finalfunc)
     {
-      sqlite3_result_error(context, "Prior Python Error", SQLITE_ERROR);
+      sqlite3_result_error(context, "Prior Python Error in step function", SQLITE_ERROR);
       goto finally;
     }
 
@@ -1368,6 +1657,15 @@ cbdispatch_final(sqlite3_context *context)
   Py_XDECREF(aggfc->stepfunc);
   Py_XDECREF(aggfc->finalfunc);
 
+  if(PyErr_Occurred() && (err_type||err_value||err_traceback))
+    {
+      PyErr_Format(PyExc_StandardError, "An exception happened during cleanup of an aggregate function, but there was already error in the step function so only that can be returned");
+      PyErr_WriteUnraisable(Py_None); /* there is no object to give, and NULL causes some versions to core dump */
+    }
+
+  if(err_type||err_value||err_traceback)
+    PyErr_Restore(err_type, err_value, err_traceback);
+
   /* sqlite3 frees the actual underlying memory we used (aggfc itself) */
 
   PyGILState_Release(gilstate);
@@ -1707,6 +2005,8 @@ static PyMethodDef Connection_methods[] = {
    "Returns the number of rows changed by last query"},
   {"totalchanges", (PyCFunction)Connection_totalchanges, METH_NOARGS, 
    "Returns the total number of changes to database since it was opened"},
+  {"getautocommit", (PyCFunction)Connection_getautocommit, METH_NOARGS, 
+   "Returns if the database is in auto-commit mode"},
   {"createcollation", (PyCFunction)Connection_createcollation, METH_VARARGS,
    "Creates a collation function"},
   {"last_insert_rowid", (PyCFunction)Connection_last_insert_rowid, METH_NOARGS,
@@ -1715,7 +2015,13 @@ static PyMethodDef Connection_methods[] = {
    "Checks if a SQL statement is complete"},
   {"setauthorizer", (PyCFunction)Connection_setauthorizer, METH_O,
    "Sets an authorizer function"},
+  {"setupdatehook", (PyCFunction)Connection_setupdatehook, METH_O,
+      "Sets an update hook"},
+  {"setrollbackhook", (PyCFunction)Connection_setrollbackhook, METH_O,
+   "Sets a callable invoked before each rollback"},
 #ifdef EXPERIMENTAL
+  {"setprofile", (PyCFunction)Connection_setprofile, METH_O,
+   "Sets a callable invoked with profile information after each statement"},
   {"setcommithook", (PyCFunction)Connection_setcommithook, METH_O,
    "Sets a callable invoked before each commit"},
   {"setprogresshandler", (PyCFunction)Connection_setprogresshandler, METH_VARARGS,
@@ -1851,7 +2157,7 @@ Cursor_dealloc(Cursor * self)
 
   if (have_error)
     {
-      /* remeber the existing error so that resetcursor won't immediately return */
+      /* remember the existing error so that resetcursor won't immediately return */
       PyErr_Fetch(&err_type, &err_value, &err_traceback);
       PyErr_Clear();
     }
@@ -1987,8 +2293,10 @@ Cursor_dobinding(Cursor *self, int arg, PyObject *obj)
 
   if(obj==Py_None)
     res=sqlite3_bind_null(self->statement, arg);
+  /* Python uses a 'long' for storage of PyInt.  This could
+     be a 32bit or 64bit quantity depending on the platform. */
   else if(PyInt_Check(obj))
-    res=sqlite3_bind_int(self->statement, arg, PyInt_AS_LONG(obj));
+    res=sqlite3_bind_int64(self->statement, arg, PyInt_AS_LONG(obj));
   else if (PyLong_Check(obj))
     /* nb: PyLong_AsLongLong can cause Python level error */
     res=sqlite3_bind_int64(self->statement, arg, PyLong_AsLongLong(obj));
@@ -2001,10 +2309,17 @@ Cursor_dobinding(Cursor *self, int arg, PyObject *obj)
         badptr=strdata;
         if(strdata)
           {
-            if(use16)
-              res=sqlite3_bind_text16(self->statement, arg, strdata, strbytes, SQLITE_TRANSIENT);
-            else
-              res=sqlite3_bind_text(self->statement, arg, strdata, strbytes, SQLITE_TRANSIENT);
+	    if(strbytes>INT32_MAX)
+	      {
+		PyErr_Format(ExcTooBig, "Unicode object is too large - SQLite only supports up to 2GB");
+	      }
+	    else
+	      {
+		if(use16)
+		  res=sqlite3_bind_text16(self->statement, arg, strdata, (int)strbytes, SQLITE_TRANSIENT);
+		else
+		  res=sqlite3_bind_text(self->statement, arg, strdata, (int)strbytes, SQLITE_TRANSIENT);
+	      }
           }
       UNIDATAEND(obj);
       if(!badptr) 
@@ -2028,10 +2343,17 @@ Cursor_dobinding(Cursor *self, int arg, PyObject *obj)
             badptr=strdata;
             if(strdata)
               {
-                if(use16)
-                  res=sqlite3_bind_text16(self->statement, arg, strdata, strbytes, SQLITE_TRANSIENT);
-                else
-                  res=sqlite3_bind_text(self->statement, arg, strdata, strbytes, SQLITE_TRANSIENT);
+		if(strbytes>INT32_MAX)
+		  {
+		    PyErr_Format(ExcTooBig, "Unicode object is too large - SQLite only supports up to 2GB");
+		  }
+		else
+		  {
+		    if(use16)
+		      res=sqlite3_bind_text16(self->statement, arg, strdata, (int)strbytes, SQLITE_TRANSIENT);
+		    else
+		      res=sqlite3_bind_text(self->statement, arg, strdata, (int)strbytes, SQLITE_TRANSIENT);
+		  }
               }
           UNIDATAEND(str2);
           Py_DECREF(str2);
@@ -2042,20 +2364,33 @@ Cursor_dobinding(Cursor *self, int arg, PyObject *obj)
             }
         }
       else
-        res=sqlite3_bind_text(self->statement, arg, val, strlen(val), SQLITE_TRANSIENT);
+	{
+	  size_t lenval=strlen(val);
+	  if(lenval>INT32_MAX)
+	      {
+		PyErr_Format(ExcTooBig, "String object is too large - SQLite only supports up to 2GB");
+		return -1;
+	      }
+	  res=sqlite3_bind_text(self->statement, arg, val, (int)lenval, SQLITE_TRANSIENT);
+	}
     }
   else if (PyBuffer_Check(obj))
     {
       const char *buffer;
-      int buflen;
+      Py_ssize_t buflen;
       if(PyObject_AsCharBuffer(obj, &buffer, &buflen))
         return -1;
-      res=sqlite3_bind_blob(self->statement, arg, buffer, buflen, SQLITE_TRANSIENT);
+      if (buflen>INT32_MAX)
+	{
+	  PyErr_Format(ExcTooBig, "Binding object is too large - SQLite only supports up to 2GB");
+	  return -1;
+	}
+      res=sqlite3_bind_blob(self->statement, arg, buffer, (int)buflen, SQLITE_TRANSIENT);
     }
   else 
     {
       PyObject *strrep=PyObject_Str(obj);
-      PyErr_Format(PyExc_TypeError, "Bad binding argument type supplied - argument #%d: %s", arg+self->bindingsoffset, strrep?PyString_AsString(strrep):"<str failed>");
+      PyErr_Format(PyExc_TypeError, "Bad binding argument type supplied - argument #%d: %s", (int)(arg+self->bindingsoffset), strrep?PyString_AsString(strrep):"<str failed>");
       Py_XDECREF(strrep);
       return -1;
     }
@@ -2096,7 +2431,8 @@ Cursor_dobindings(Cursor *self)
               return -1;
             }
 
-          key++; /* first char is a colon which we skip */
+	  assert(*key==':' || *key=='$');
+          key++; /* first char is a colon or dollar which we skip */
 
           for(chk=key;*chk && !((*chk)&0x80); chk++);
           if(*chk)
@@ -2131,14 +2467,14 @@ Cursor_dobindings(Cursor *self)
   if(*self->zsqlnextpos && sz-self->bindingsoffset<nargs)
     {
       PyErr_Format(ExcBindings, "Incorrect number of bindings supplied.  The current statement uses %d and there are only %d left.  Current offset is %d",
-                   nargs, (self->bindings)?sz:0, self->bindingsoffset);
+                   nargs, (self->bindings)?sz:0, (int)(self->bindingsoffset));
       return -1;
     }
   /* no more statements */
   if(!*self->zsqlnextpos && sz-self->bindingsoffset!=nargs)
     {
       PyErr_Format(ExcBindings, "Incorrect number of bindings supplied.  The current statement uses %d and there are %d supplied.  Current offset is %d",
-                   nargs, (self->bindings)?sz:0, self->bindingsoffset);
+                   nargs, (self->bindings)?sz:0, (int)(self->bindingsoffset));
       return -1;
     }
   
@@ -2162,7 +2498,7 @@ Cursor_dobindings(Cursor *self)
 
 typedef struct { 
   const char *previouszsqlpos;  /* where the begining of the statement was */
-  int savedbindingsoffset;      /* where the bindings began */
+  Py_ssize_t savedbindingsoffset;      /* where the bindings began */
 } exectrace_oldstate;
   
 static int
@@ -2271,7 +2607,7 @@ Cursor_step(Cursor *self)
         {
         case SQLITE_ROW:
           self->status=C_ROW;
-          return (PyObject*)self;
+          return (PyErr_Occurred())?(NULL):((PyObject*)self);
         case SQLITE_BUSY:
           self->status=C_BEGIN;
           SET_EXC(self->connection->db,res);
@@ -2296,6 +2632,11 @@ Cursor_step(Cursor *self)
           return NULL;
 
         case SQLITE_DONE:
+	  if (PyErr_Occurred())
+	    {
+	      self->status=C_DONE;
+	      return NULL;
+	    }
           break;
           
         }
@@ -2621,12 +2962,12 @@ Cursor_next(Cursor *self)
 
         case SQLITE_BLOB:
           {
-            int sz=sqlite3_column_bytes(self->statement, i);
+            Py_ssize_t sz=sqlite3_column_bytes(self->statement, i);
             item=PyBuffer_New(sz);
             if(item)
               {
                 void *buffy=0;
-                int sz2=sz;
+                Py_ssize_t sz2=sz;
                 if(!PyObject_AsWriteBuffer(item, &buffy, &sz2))
                   memcpy(buffy, sqlite3_column_blob(self->statement, i), sz);
                 else
@@ -2816,12 +3157,30 @@ getapswversion(void)
   return PyString_FromString(APSW_VERSION);
 }
 
+static PyObject *
+enablesharedcache(PyObject *self, PyObject *args)
+{
+  int setting,res;
+  if(!PyArg_ParseTuple(args, "i:enablesharedcache(boolean)", &setting))
+    return NULL;
+
+  res=sqlite3_enable_shared_cache(setting);
+  SET_EXC(NULL, res);
+
+  if(res!=SQLITE_OK)
+    return NULL;
+
+  return Py_BuildValue("");
+}
 
 static PyMethodDef module_methods[] = {
   {"sqlitelibversion", (PyCFunction)getsqliteversion, METH_NOARGS,
    "Return the version of the SQLite library"},
   {"apswversion", (PyCFunction)getapswversion, METH_NOARGS,
    "Return the version of the APSW wrapper"},
+  {"enablesharedcache", (PyCFunction)enablesharedcache, METH_VARARGS,
+   "Sets shared cache semantics for this thread"},
+
     {NULL}  /* Sentinel */
 };
 
@@ -2835,8 +3194,8 @@ initapsw(void)
 {
     PyObject* m;
 
-    assert(sizeof(int)==sizeof(long));  /* Python treats PyIntObjects as long */
     assert(sizeof(int)==4);             /* we expect 32 bit ints */
+    assert(sizeof(long long)==8);             /* we expect 64 bit long long */
 
     if (PyType_Ready(&ConnectionType) < 0)
         return;
@@ -2905,6 +3264,13 @@ initapsw(void)
 #endif
 #ifdef SQLITE_REINDEX
     ADDINT(SQLITE_REINDEX);
+#endif
+    /* in 3.2 */
+#ifdef SQLITE_COPY
+    ADDINT(SQLITE_COPY);
+#endif
+#ifdef SQLITE_ANALYZE
+    ADDINT(SQLITE_ANALYZE);
 #endif
 }
 
